@@ -27,8 +27,9 @@ flagReader.py options are still supported for backward compatibility
 """
 import cmd, re, os, sys, optparse, subprocess, tempfile, pyparsing, doctest
 import unittest, string, datetime, urllib
+from code import InteractiveConsole, InteractiveInterpreter, softspace
 from optparse import make_option
-__version__ = '0.4.8'
+__version__ = '0.5.1'
 
 class OptionParser(optparse.OptionParser):
     def exit(self, status=0, msg=None):
@@ -220,21 +221,53 @@ def replace_with_file_contents(fname):
             result = '< %s' % fname[0]  # wasn't a file after all
     else:
         result = getPasteBuffer()
-    return result
+    return result      
+
+class EmbeddedConsoleExit(Exception):
+    pass
+
+class MyInteractiveConsole(InteractiveConsole):
+    def runcode(self, code):
+        """Execute a code object.
+
+        When an exception occurs, self.showtraceback() is called to
+        display a traceback.  All exceptions are caught except
+        SystemExit, which is reraised.
+
+        A note about KeyboardInterrupt: this exception may occur
+        elsewhere in this code, and may not always be caught.  The
+        caller should be prepared to deal with it.
         
+        Copied directly from code.InteractiveInterpreter, except for
+        EmbeddedConsoleExit exceptions.
+        """
+        try:
+            exec code in self.locals
+        except (SystemExit, EmbeddedConsoleExit):
+            raise
+        except:
+            self.showtraceback()
+        else:
+            if softspace(sys.stdout, 0):
+                print
+
 class Cmd(cmd.Cmd):
     echo = False
     case_insensitive = True
     continuation_prompt = '> '  
     timing = False
     legalChars = '!#$%.:?@_' + pyparsing.alphanums + pyparsing.alphas8bit  # make sure your terminators are not in here!
-    shortcuts = {'?': 'help', '!': 'shell', '@': 'load' }
+    shortcuts = {'?': 'help', '!': 'shell', '@': 'load', '@@': '_relative_load'}
     excludeFromHistory = '''run r list l history hi ed edit li eof'''.split()
     noSpecialParse = 'set ed edit exit'.split()
     defaultExtension = 'txt'
     default_file_name = 'command.txt'
-    settable = ['prompt', 'continuation_prompt', 'default_file_name', 'editor', 'case_insensitive', 
-                'echo', 'timing']
+    abbrev = True
+    nonpythoncommand = 'cmd'
+    current_script_dir = None
+    reserved_words = []
+    settable = ['prompt', 'continuation_prompt', 'default_file_name', 'editor',
+                'case_insensitive', 'echo', 'timing', 'abbrev']   
     settable.sort()
     
     editor = os.environ.get('EDITOR')
@@ -252,32 +285,32 @@ class Cmd(cmd.Cmd):
         self.stdout.write("""
         Commands are %(casesensitive)scase-sensitive.
         Commands may be terminated with: %(terminators)s
-        Settable parameters: %(settable)s
-        """ % 
+        Settable parameters: %(settable)s\n""" % \
         { 'casesensitive': (self.case_insensitive and 'not ') or '',
           'terminators': str(self.terminators),
           'settable': ' '.join(self.settable)
         })
         
     def do_help(self, arg):
-        try:
-            fn = getattr(self, 'do_' + arg)
-            if fn and fn.optionParser:
+        funcname = self.func_named(arg)
+        if funcname:
+            fn = getattr(self, funcname)
+            try:
                 fn.optionParser.print_help(file=self.stdout)
-                return
-        except AttributeError:
-            pass
-        cmd.Cmd.do_help(self, arg)
+            except AttributeError:
+                cmd.Cmd.do_help(self, funcname[3:])
         
     def __init__(self, *args, **kwargs):        
         cmd.Cmd.__init__(self, *args, **kwargs)
         self.history = History()
         self._init_parser()
         self.pystate = {}
-        
+        self.shortcuts = sorted(self.shortcuts.items(), reverse=True)
+        self.keywords = self.reserved_words + [fname[3:] for fname in dir(self) 
+                                               if fname.startswith('do_')]        
     def do_shortcuts(self, args):
         """Lists single-key shortcuts available."""
-        result = "\n".join('%s: %s' % (sc[0], sc[1]) for sc in self.shortcuts.items())
+        result = "\n".join('%s: %s' % (sc[0], sc[1]) for sc in sorted(self.shortcuts))
         self.stdout.write("Single-key shortcuts for other commands:\n%s\n" % (result))
 
     prefixParser = pyparsing.Empty()
@@ -469,6 +502,7 @@ class Cmd(cmd.Cmd):
             self.blankLineTerminationParser = ((self.multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(self.blankLineTerminator).setParseAction(lambda x: x[0].strip())('args') + self.blankLineTerminator)('statement')
         self.multilineParser = (((self.multilineCommand ^ oneLineCommand) + SkipToLast(terminatorParser).setParseAction(lambda x: x[0].strip())('args') + terminatorParser)('statement') +
                                 pyparsing.SkipTo(outputParser ^ pipe ^ stringEnd).setParseAction(lambda x: x[0].strip())('suffix') + afterElements)
+        self.multilineParser.ignore(self.commentInProgress)
         self.singleLineParser = ((oneLineCommand + pyparsing.SkipTo(terminatorParser ^ stringEnd ^ pipe ^ outputParser).setParseAction(lambda x:x[0].strip())('args'))('statement') +
                                  pyparsing.Optional(terminatorParser) + afterElements)
         #self.multilineParser = self.multilineParser.setResultsName('multilineParser')
@@ -481,7 +515,7 @@ class Cmd(cmd.Cmd):
             self.prefixParser + self.blankLineTerminationParser | 
             self.prefixParser + self.multilineCommand + pyparsing.SkipTo(stringEnd)
             )
-        self.parser.ignore(pyparsing.quotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)
+        self.parser.ignore(pyparsing.quotedString).ignore(self.commentGrammars)
         
         inputMark = pyparsing.Literal('<')
         inputMark.setParseAction(lambda x: '')
@@ -503,7 +537,7 @@ class Cmd(cmd.Cmd):
         else:
             raw = self.preparse(raw, **kwargs)
             s = self.inputParser.transformString(raw.lstrip())
-            for (shortcut, expansion) in self.shortcuts.items():
+            for (shortcut, expansion) in self.shortcuts:
                 if s.lower().startswith(shortcut):
                     s = s.replace(shortcut, expansion + ' ', 1)
                     break
@@ -522,8 +556,20 @@ class Cmd(cmd.Cmd):
     def postparsing_precmd(self, statement):
         stop = 0
         return stop, statement
+
     def postparsing_postcmd(self, stop):
         return stop
+    def func_named(self, arg):
+        result = None
+        target = 'do_' + arg
+        if target in dir(self):
+            result = target
+        else:
+            if self.abbrev:   # accept shortened versions of commands
+                funcs = [fname for fname in self.keywords if fname.startswith(arg)]
+                if len(funcs) == 1:
+                    result = 'do_' + funcs[0]
+        return result
     def onecmd(self, line):
         """Interpret the argument as though it had been typed in response
         to the prompt.
@@ -549,7 +595,8 @@ class Cmd(cmd.Cmd):
         except Exception, e:
             print e
             return 0
-
+        if statement.parsed.command not in self.excludeFromHistory:
+            self.history.append(statement.parsed.raw)
         try:
             (stop, statement) = self.postparsing_precmd(statement)
         except Exception, e:
@@ -585,20 +632,20 @@ class Cmd(cmd.Cmd):
                     self.stdout.write(getPasteBuffer())
         try:
             # "heart" of the command, replace's cmd's onecmd()
-            self.lastcmd = statement.parsed.expanded
+            self.lastcmd = statement.parsed.expanded   
+            funcname = self.func_named(statement.parsed.command)
+            if not funcname:
+                return self.postparsing_postcmd(self.default(statement))  
             try:
-                func = getattr(self, 'do_' + statement.parsed.command)
+                func = getattr(self, funcname)
             except AttributeError:
-                return self.postparsing_postcmd(self.default(statement))
+                return self.postparsing_postcmd(self.default(statement))                  
             timestart = datetime.datetime.now()
             stop = func(statement) 
             if self.timing:
                 print 'Elapsed: %s' % str(datetime.datetime.now() - timestart)
         except Exception, e:
             print e
-        try:
-            if statement.parsed.command not in self.excludeFromHistory:
-                self.history.append(statement.parsed.raw)
         finally:
             if statekeeper:
                 if statement.parsed.output and not statement.parsed.outputTo:
@@ -741,40 +788,44 @@ class Cmd(cmd.Cmd):
         'execute a command as if at the OS prompt.'
         os.system(arg)
         
+    def _attempt_py_command(self, arg):
+        try:
+            result = eval(arg, self.pystate)
+            print repr(result)
+        except SyntaxError:
+            exec(arg, self.pystate)       
+        return
+        
     def do_py(self, arg):  
         '''
         py <command>: Executes a Python command.
-        py: Enters interactive Python mode (end with `end py`).
+        py: Enters interactive Python mode.
+        End with `Ctrl-D` (Unix) / `Ctrl-Z` (Windows), `quit()`, 'exit()`.
+        Non-python commands can be issued with `cmd("your command")`.
         '''
         if arg.strip():
-            try:
-                result = eval(arg, self.pystate)
-                print repr(result)
-            except SyntaxError:
-                try:
-                    exec(arg, self.pystate)
-                except Exception:
-                    raise
-            except Exception, e:
-                print e        
+            interp = InteractiveInterpreter(locals=self.pystate)
+            interp.runcode(arg)
         else:
-            print 'Now accepting python commands; end with `end py`'
-            buffer = [self.pseudo_raw_input('>>> ')]
-            while buffer[-1].lower().split()[:2] != ['end','py']:
-                try:
-                    buf = '\n'.join(buffer)
-                    try:
-                        result = eval(buf, self.pystate)
-                        print repr(result)
-                    except SyntaxError:
-                        exec(buf, self.pystate)
-                    buffer = [self.pseudo_raw_input('>>> ')]
-                except SyntaxError:
-                    buffer.append(self.pseudo_raw_input('... '))
-                except Exception, e:
-                    print e
-                    buffer = [self.pseudo_raw_input('>>> ')]
-
+            interp = MyInteractiveConsole(locals=self.pystate)
+            def quit():
+                raise EmbeddedConsoleExit
+            def onecmd(arg):
+                return self.onecmd(arg + '\n')
+            self.pystate['quit'] = quit
+            self.pystate['exit'] = quit
+            self.pystate[self.nonpythoncommand] = onecmd
+            try:
+                cprt = 'Type "help", "copyright", "credits" or "license" for more information.'        
+                keepstate = Statekeeper(sys, ('stdin','stdout'))
+                sys.stdout = self.stdout
+                sys.stdin = self.stdin
+                interp.interact(banner= "Python %s on %s\n%s\n(%s)\n%s" %
+                       (sys.version, sys.platform, cprt, self.__class__.__name__, self.do_py.__doc__))
+            except EmbeddedConsoleExit:
+                pass
+            keepstate.restore()
+            
     def do_history(self, arg):
         """history [arg]: lists past commands issued
         
@@ -873,34 +924,52 @@ class Cmd(cmd.Cmd):
         except Exception, e:
             print 'Error saving %s: %s' % (fname, str(e))
             
-    urlre = re.compile('(https?://[-\\w\\./]+)')
-    def do_load(self, fname=None):           
-        """Runs script of command(s) from a file or URL."""
-        if fname is None:
-            fname = self.default_file_name
-        #keepstate = Statekeeper(self, ('stdin','use_rawinput','prompt','continuation_prompt'))
-        keepstate = Statekeeper(self, ('stdin','use_rawinput','continuation_prompt'))
-        try:
-            if isinstance(fname, file):
-                target = open(fname, 'r')
+    def read_file_or_url(self, fname):
+        if isinstance(fname, file):
+            target = open(fname, 'r')
+        else:
+            match = self.urlre.match(fname)
+            if match:
+                target = urllib.urlopen(match.group(1))
             else:
-                match = self.urlre.match(fname)
-                if match:
-                    target = urllib.urlopen(match.group(1))
-                else:
-                    fname = os.path.expanduser(fname)
-                    try:
-                        target = open(os.path.expanduser(fname), 'r')
-                    except IOError, e:                    
-                        target = open('%s.%s' % (os.path.expanduser(fname), 
-                                                 self.defaultExtension), 'r')
+                fname = os.path.expanduser(fname)
+                try:
+                    result = open(os.path.expanduser(fname), 'r')
+                except IOError, e:                    
+                    result = open('%s.%s' % (os.path.expanduser(fname), 
+                                             self.defaultExtension), 'r')
+        return result
+        
+    def do__relative_load(self, arg=None):
+        '''
+        Runs commands in script at file or URL; if this is called from within an
+        already-running script, the filename will be interpreted relative to the 
+        already-running script's directory.'''
+        if arg:
+            arg = arg.split(None, 1)
+            targetname, args = arg[0], (arg[1:] or [''])[0]
+            targetname = os.path.join(self.current_script_dir or '', targetname)
+            self.do__load('%s %s' % (targetname, args))
+    
+    urlre = re.compile('(https?://[-\\w\\./]+)')
+    def do_load(self, arg=None):           
+        """Runs script of command(s) from a file or URL."""
+        if arg is None:
+            targetname = self.default_file_name
+        else:
+            arg = arg.split(None, 1)
+            targetname, args = arg[0], (arg[1:] or [''])[0].strip()
+        try:
+            target = self.read_file_or_url(targetname)
         except IOError, e:
-            print 'Problem accessing script from %s: \n%s' % (fname, e)
-            keepstate.restore()
+            print 'Problem accessing script from %s: \n%s' % (targetname, e)
             return
+        keepstate = Statekeeper(self, ('stdin','use_rawinput','prompt',
+                                       'continuation_prompt','current_script_dir'))
         self.stdin = target    
         self.use_rawinput = False
         self.prompt = self.continuation_prompt = ''
+        self.current_script_dir = os.path.split(targetname)[0]
         stop = self.cmdloop()
         self.stdin.close()
         keepstate.restore()
